@@ -3,7 +3,7 @@
 ## build_info.py
 ##
 ## Author:   Toke Høiland-Jørgensen (toke@toke.dk)
-## Date:      6 december 2012
+## Date:      6 December 2012
 ## Copyright (c) 2012, Toke Høiland-Jørgensen
 ##
 ## This program is free software: you can redistribute it and/or modify
@@ -24,5 +24,13 @@
 import os
 
 # this value works for the source distribution
-VERSION="0.6.1"
+VERSION="0.7.0"
 DATA_DIR=os.path.normpath(os.path.join(os.path.dirname(__file__), '..'))
+
+if VERSION.endswith("-git") and os.path.exists(os.path.join(DATA_DIR, '.git')):
+    try:
+        import subprocess
+        commit = subprocess.check_output(["git", "log", "--format=%h", "-1"], cwd=DATA_DIR).decode()
+        VERSION+="-%s" % commit.strip()
+    except:
+        pass
