@@ -3,7 +3,7 @@
 ## resultset.py
 ##
 ## Author:   Toke Høiland-Jørgensen (toke@toke.dk)
-## Date:     24 november 2012
+## Date:     24 November 2012
 ## Copyright (c) 2012, Toke Høiland-Jørgensen
 ##
 ## This program is free software: you can redistribute it and/or modify
@@ -21,6 +21,9 @@
 
 import json, os, math, re, sys
 from datetime import datetime
+from calendar import timegm
+from itertools import repeat
+from copy import deepcopy
 
 try:
     from dateutil.parser import parse as parse_date
@@ -37,28 +40,76 @@ from .util import gzip_open
 # Controls pretty-printing of json dumps
 JSON_INDENT=None
 
-__all__ = ['ResultSet']
+__all__ = ['new', 'load']
+
+RECORDED_SETTINGS = (
+    "NAME",
+    "HOST",
+    "HOSTS",
+    "TIME",
+    "LOCAL_HOST",
+    "TITLE",
+    "NOTE",
+    "LENGTH",
+    "TOTAL_LENGTH",
+    "STEP_SIZE",
+    "TEST_PARAMETERS",
+    "NETPERF_WRAPPER_VERSION",
+    "IP_VERSION",
+    "BATCH_NAME",
+    "BATCH_TIME",
+    "DATA_FILENAME",
+    "HTTP_GETTER_URLLIST",
+    "HTTP_GETTER_DNS",
+    "HTTP_GETTER_WORKERS",
+    )
+
+# Time settings will be serialised as ISO timestamps and stored in memory as
+# datetime instances
+TIME_SETTINGS = ("TIME", "BATCH_TIME", "T0")
+
+def new(settings):
+    d = {}
+    for a in RECORDED_SETTINGS:
+        d[a] = deepcopy(getattr(settings,a,None))
+    return ResultSet(**d)
+
+def load(filename, absolute=False):
+    return ResultSet.load_file(filename, absolute)
 
 class ResultSet(object):
+    SUFFIX = '.json.gz'
     def __init__(self, **kwargs):
-        self.metadata = kwargs
-        if not 'TIME' in self.metadata:
-            self.metadata['TIME'] = datetime.now()
-        if not 'NAME' in self.metadata:
-            raise RuntimeError("Missing name for resultset")
         self._x_values = []
         self._results = OrderedDict()
+        self._filename = None
+        self._absolute = False
+        self.metadata = kwargs
+        if not 'TIME' in self.metadata or self.metadata['TIME'] is None:
+            self.metadata['TIME'] = datetime.now()
+        if not 'NAME' in self.metadata or self.metadata['NAME'] is None:
+            raise RuntimeError("Missing name for resultset")
+        if not 'DATA_FILENAME' in self.metadata or self.metadata['DATA_FILENAME'] is None:
+            self.metadata['DATA_FILENAME'] = self.dump_file
+        if not self.metadata['DATA_FILENAME'].endswith(self.SUFFIX):
+            self.metadata['DATA_FILENAME'] += self.SUFFIX
+        self._filename = self.metadata['DATA_FILENAME']
 
-    def meta(self, k=None):
+    def meta(self, k=None, v=None):
         if k:
+            if v:
+                self.metadata[k] = v
             return self.metadata[k]
         return self.metadata
 
+    def label(self):
+        return self.metadata["TITLE"] or self.metadata["TIME"].strftime("%Y-%m-%d %H:%M:%S")
+
     def get_x_values(self):
         return self._x_values
     def set_x_values(self, x_values):
         assert not self._x_values
-        self._x_values = x_values
+        self._x_values = list(x_values)
     x_values = property(get_x_values, set_x_values)
 
     def add_result(self, name, data):
@@ -87,6 +138,21 @@ class ResultSet(object):
         if data:
             raise RuntimeError("Unexpected data point(s): %s" % list(data.keys()))
 
+    def concatenate(self, res):
+        if self._absolute:
+            x0 = 0.0
+            # When concatenating using absolute values, insert an empty data
+            # point midway between the data series, to prevent the lines for
+            # each distinct data series from being joined together.
+            xnext = (self.x_values[-1] + res.x_values[0])/2.0
+            self.append_datapoint(xnext, zip(res.series_names, repeat(None)))
+        else:
+            x0 = self.x_values[-1] + self.meta("STEP_SIZE")
+        for point in res:
+            x = point[0] + x0
+            data = dict(zip(res.series_names, point[1:]))
+            self.append_datapoint(x, data)
+
     def last_datapoint(self, series):
         data = self.series(series)
         if not data:
@@ -101,6 +167,12 @@ class ResultSet(object):
             return self.smoothed(name, smooth)
         return self._results[name]
 
+    def __getitem__(self, name):
+        return self.series(name)
+
+    def __contains__(self, name):
+        return name in self._results
+
     def smoothed(self, name, amount):
         res = self._results[name]
         smooth_res = []
@@ -123,7 +195,13 @@ class ResultSet(object):
         if keys is None:
             keys = self.series_names
         for i in range(len(self._x_values)):
-            yield [self._x_values[i]]+[self._results[k][i] for k in keys]
+            y = [self._x_values[i]]
+            for k in keys:
+                if k in self._results:
+                    y.append(self._results[k][i])
+                else:
+                    y.append(None)
+            yield y
 
     def __iter__(self):
         return self.zipped()
@@ -131,9 +209,15 @@ class ResultSet(object):
     def __len__(self):
         return len(self._x_values)
 
+    def serialise_metadata(self):
+        metadata = self.metadata.copy()
+        for t in TIME_SETTINGS:
+            if t in metadata and metadata[t] is not None:
+                metadata[t] = metadata[t].isoformat()
+        return metadata
+
     def serialise(self):
-        metadata = dict(self.metadata)
-        metadata['TIME'] = metadata['TIME'].isoformat()
+        metadata = self.serialise_metadata()
         return {
             'metadata': metadata,
             'x_values': self._x_values,
@@ -161,12 +245,15 @@ class ResultSet(object):
         return self._gen_filename()
 
     def _gen_filename(self):
+        if self._filename is not None:
+            return self._filename
         if 'TITLE' in self.metadata and self.metadata['TITLE']:
-            return "%s-%s.%s.json.gz" % (self.metadata['NAME'],
+            return "%s-%s.%s%s" % (self.metadata['NAME'],
                                          self.metadata['TIME'].isoformat().replace(":", ""),
-                                         re.sub("[^A-Za-z0-9]", "_", self.metadata['TITLE'])[:50])
+                                         re.sub("[^A-Za-z0-9]", "_", self.metadata['TITLE'])[:50],
+                                         self.SUFFIX)
         else:
-            return "%s-%s.json.gz" % (self.metadata['NAME'], self.metadata['TIME'].isoformat().replace(":", ""))
+            return "%s-%s%s" % (self.metadata['NAME'], self.metadata['TIME'].isoformat().replace(":", ""), self.SUFFIX)
 
     def dump_dir(self, dirname):
         self._dump_file = os.path.join(dirname, self._gen_filename())
@@ -181,32 +268,39 @@ class ResultSet(object):
             self._dump_file = None
 
     @classmethod
-    def unserialise(cls, obj):
+    def unserialise(cls, obj, absolute=False):
         metadata = dict(obj['metadata'])
-        if 'TIME' in metadata:
-            metadata['TIME'] = parse_date(metadata['TIME'])
+        for t in TIME_SETTINGS:
+            if t in metadata and metadata[t] is not None:
+                metadata[t] = parse_date(metadata[t])
         rset = cls(**metadata)
-        rset.x_values = obj['x_values']
+        if absolute:
+            t0 = metadata.get('T0', metadata.get('TIME'))
+            x0 = timegm(t0.timetuple()) + t0.microsecond / 1000000.0
+            rset.x_values = [x+x0 for x in obj['x_values']]
+            rset._absolute = True
+        else:
+            rset.x_values = obj['x_values']
         for k,v in list(obj['results'].items()):
             rset.add_result(k,v)
         return rset
 
     @classmethod
-    def load(cls, fp):
-        obj = cls.unserialise(json.load(fp))
+    def load(cls, fp, absolute=False):
+        obj = cls.unserialise(json.load(fp), absolute)
         if hasattr(fp, 'name'):
             obj._dump_file = fp.name
         return obj
 
     @classmethod
-    def load_file(cls, filename):
+    def load_file(cls, filename, absolute=False):
         try:
             if filename.endswith(".gz"):
                 o = gzip_open
             else:
                 o = open
             fp = o(filename, 'rt')
-            r = cls.load(fp)
+            r = cls.load(fp, absolute)
             fp.close()
             return r
         except IOError:
