@@ -3,7 +3,7 @@
 ## runners.py
 ##
 ## Author:   Toke Høiland-Jørgensen (toke@toke.dk)
-## Date:     16 oktober 2012
+## Date:     16 October 2012
 ## Copyright (c) 2012, Toke Høiland-Jørgensen
 ##
 ## This program is free software: you can redistribute it and/or modify
@@ -19,18 +19,44 @@
 ## You should have received a copy of the GNU General Public License
 ## along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import threading, time, shlex, subprocess, re, time, sys, math, os, tempfile, signal
+import threading, time, shlex, subprocess, re, time, sys, math, os, tempfile, \
+  signal, hmac, hashlib, calendar, socket
 
 from datetime import datetime
 
-from .settings import settings, Glob
+from .util import classname
+
+try:
+    from defusedxml.xmlrpc import monkey_patch
+    monkey_patch()
+    del monkey_patch
+    XML_DEFUSED=True
+except ImportError:
+    XML_DEFUSED=False
+
+try:
+    # python 2
+    import xmlrpclib as xmlrpc
+except ImportError:
+    import xmlrpc.client as xmlrpc
+
+
+from .util import Glob
+
+def get(name):
+    cname = classname(name, "Runner")
+    if not cname in globals():
+        raise RuntimeError("Runner not found: '%s'" % name)
+    return globals()[cname]
 
 class ProcessRunner(threading.Thread):
     """Default process runner for any process."""
+    silent = False
 
-    def __init__(self, name, command, delay, *args, **kwargs):
+    def __init__(self, name, settings, command, delay, *args, **kwargs):
         threading.Thread.__init__(self)
         self.name = name
+        self.settings = settings
         self.command = command
         self.args = shlex.split(self.command)
         self.delay = delay
@@ -38,12 +64,16 @@ class ProcessRunner(threading.Thread):
         self.killed = False
         self.pid = None
         self.returncode = None
+        self.kill_lock = threading.Lock()
+        self.metadata = {}
+        self.out = ""
+        self.err = ""
 
     def fork(self):
         # Use named temporary files to avoid errors on double-delete when
         # running on Windows/cygwin.
-        self.stdout = tempfile.NamedTemporaryFile(delete=False)
-        self.stderr = tempfile.NamedTemporaryFile(delete=False)
+        self.stdout = tempfile.NamedTemporaryFile(prefix="netperf-wrapper-", delete=False)
+        self.stderr = tempfile.NamedTemporaryFile(prefix="netperf-wrapper-", delete=False)
 
         pid = os.fork()
 
@@ -61,14 +91,17 @@ class ProcessRunner(threading.Thread):
             self.pid = pid
 
     def kill(self):
-        if self.killed:
-            return
+        with self.kill_lock:
+            self.killed = True
         if self.pid is not None:
             try:
-                os.kill(self.pid, signal.SIGINT)
+                os.kill(self.pid, signal.SIGTERM)
             except OSError:
                 pass
-        self.killed = True
+
+    def is_killed(self):
+        with self.kill_lock:
+            return self.killed
 
     # helper function from subprocess module
     def _handle_exitstatus(self, sts, _WIFSIGNALED=os.WIFSIGNALED,
@@ -97,7 +130,7 @@ class ProcessRunner(threading.Thread):
         self._handle_exitstatus(sts)
 
         self.stdout.seek(0)
-        self.out = self.stdout.read().decode()
+        self.out += self.stdout.read().decode()
         try:
             # Close and remove the temporary file. This might fail, but we're going
             # to assume that is okay.
@@ -108,7 +141,7 @@ class ProcessRunner(threading.Thread):
             pass
 
         self.stderr.seek(0)
-        self.err = self.stderr.read().decode()
+        self.err += self.stderr.read().decode()
         try:
             filename = self.stderr.name
             self.stderr.close()
@@ -116,18 +149,21 @@ class ProcessRunner(threading.Thread):
         except OSError:
             pass
 
-        if self.killed:
-            return
+        # Even with locking, kill detection is not reliable; sleeping seems to
+        # help. *sigh* -- threading.
+        time.sleep(0.2)
 
-        if self.returncode:
-            sys.stderr.write("Warning: Program exited non-zero.\nCommand: %s\n" % self.command)
+        if self.is_killed():
+            return
+        elif self.returncode:
+            sys.stderr.write("Warning: Program exited non-zero (%d).\nCommand: %s\n" % (self.returncode, self.command))
             sys.stderr.write("Program output:\n")
             sys.stderr.write("  " + "\n  ".join(self.err.splitlines()) + "\n")
             sys.stderr.write("  " + "\n  ".join(self.out.splitlines()) + "\n")
             self.result = None
         else:
             self.result = self.parse(self.out)
-            if not self.result:
+            if not self.result and not self.silent:
                 sys.stderr.write("Warning: Command produced no valid data.\n"
                                  "Data series: %s\n"
                                  "Runner: %s\n"
@@ -138,12 +174,98 @@ class ProcessRunner(threading.Thread):
 
     def parse(self, output):
         """Default parser returns the last (whitespace-separated) word of
-        output."""
+        output as a float."""
 
-        return output.split()[-1].strip()
+        return float(output.split()[-1].strip())
 
 DefaultRunner = ProcessRunner
 
+class SilentProcessRunner(ProcessRunner):
+    silent = True
+    def parse(self, output):
+        return None
+
+class DitgRunner(ProcessRunner):
+    """Runner for D-ITG with a control server."""
+
+    def __init__(self, name, settings, duration, interval, **kwargs):
+        ProcessRunner.__init__(self, name, settings, **kwargs)
+        host = self.settings.DITG_CONTROL_HOST or self.settings.HOST
+        self.proxy = xmlrpc.ServerProxy("http://%s:%s" % (host, self.settings.DITG_CONTROL_PORT),
+                                        allow_none=True)
+        self.ditg_secret = self.settings.DITG_CONTROL_SECRET
+        self.duration = duration
+        self.interval = interval
+
+
+    def start(self):
+        try:
+            interval = int(self.interval*1000)
+            hm = hmac.new(self.ditg_secret.encode(), digestmod=hashlib.sha256)
+            hm.update(str(self.duration).encode())
+            hm.update(str(interval).encode())
+            params = self.proxy.request_new_test(self.duration, interval, hm.hexdigest())
+            if params['status'] != 'OK':
+                if 'message' in params:
+                    raise RuntimeError("Unable to request D-ITG test. Control server reported error: %s" % params['message'])
+                else:
+                    raise RuntimeError("Unable to request D-ITG test. Control server reported an unspecified error.")
+            self.test_id = params['test_id']
+            self.out += "Test ID: %s\n" % self.test_id
+        except (xmlrpc.Fault, socket.error) as e:
+            raise RuntimeError("Error while requesting D-ITG test: %s" % e)
+        self.command = self.command.format(signal_port = params['port'], dest_port=params['port']+1)
+        self.args = shlex.split(self.command)
+        ProcessRunner.start(self)
+
+    def parse(self, output):
+        data = ""
+        utc_offset = 0
+        results = {}
+        try:
+            # The control server has a grace period after the test ends, so we
+            # don't know exactly when the test results are going to be ready. We
+            # assume that it will be within ten seconds.
+            for i in range(10):
+                res = self.proxy.get_test_results(self.test_id)
+                if res['status'] == 'OK':
+                    data = res['data']
+                    self.out += data
+                    utc_offset = res['utc_offset']
+                    break
+                time.sleep(1)
+            if res['status'] != 'OK':
+                if 'message' in res:
+                    self.err += "Error while getting results. Control server reported error: %s.\n" % res['message']
+                else:
+                    self.err += "Error while getting results. Control server reported unknown error.\n"
+        except xmlrpc.Fault as e:
+            self.err += "Error while getting results: %s.\n" % e
+
+        now = time.time()
+        tzoffset = (datetime.fromtimestamp(now) - datetime.utcfromtimestamp(now)).seconds
+        offset = utc_offset + tzoffset
+
+        # D-ITG *should* output about 50 bytes of data per data point. However,
+        # sometimes it runs amok and outputs megabytes of erroneous data. So, if
+        # the length of the data is more than ten times the expected value,
+        # abort rather than try to process the data.
+        if len(data) > (self.duration/self.interval) * 500:
+            self.err += "D-ITG output too much data (%d bytes).\n" % len(data)
+            return results
+
+        for line in data.splitlines():
+            if not line.strip():
+                continue
+            parts = [float(i) for i in line.split()]
+            timestamp = parts.pop(0) + offset
+            for i,n in enumerate(('bitrate', 'delay', 'jitter', 'loss')):
+                if not n in results:
+                    results[n] = []
+                results[n].append([timestamp, parts[i]])
+
+        return results
+
 class NetperfDemoRunner(ProcessRunner):
     """Runner for netperf demo mode."""
 
@@ -152,32 +274,64 @@ class NetperfDemoRunner(ProcessRunner):
         pairs."""
 
         result = []
-        lines = output.split("\n")
+        lines = output.strip().splitlines()
+        avg_dur = None
+        alpha = 0.5
         for line in lines:
             if line.startswith("Interim"):
                 parts = line.split()
-                result.append([float(parts[9]), float(parts[2])])
+
+                # Calculate an EWMA of the netperf sampling duration and exclude
+                # data points from a sampling period that is more than an order
+                # of magnitude higher or lower than this average; these are
+                # probably the result of netperf spitting out a measurement at
+                # the end of a run after having lost the measurement flow during
+                # the run, or a very short interval giving a very high bandwidth
+                # measurement
+                dur = float(parts[5])
+                if avg_dur is None:
+                    avg_dur = dur
+
+                if dur < avg_dur * 10.0 and dur > avg_dur / 10.0:
+                    result.append([float(parts[9]), float(parts[2])])
+                    avg_dur = alpha * avg_dur + (1.0-alpha) * dur
+        try:
+            self.metadata['MEAN_VALUE'] = float(lines[-1])
+        except ValueError:
+            pass
 
         return result
 
-class PingRunner(ProcessRunner):
-    """Runner for ping/ping6 in timestamped (-D) mode."""
+class RegexpRunner(ProcessRunner):
 
-    pingline_regex = re.compile(r'^\[([0-9]+\.[0-9]+)\].*time=([0-9]+(?:\.[0-9]+)?) ms$')
-    fpingline_regex = re.compile(r'^\[([0-9]+\.[0-9]+)\].*:.*, ([0-9]+(?:\.[0-9]+)?) ms \(.*\)$')
+    """Runner that matches each line to one or more regular expressions,
+    returning the values from the first matched.
+
+    The regular expressions must define symbolic groups 'time' and 'value'."""
+
+    regexes = []
 
     def parse(self, output):
         result = []
         lines = output.split("\n")
         for line in lines:
-            match = self.pingline_regex.match(line)
-            if not match:
-                match = self.fpingline_regex.match(line)
-            if match:
-                result.append([float(match.group(1)), float(match.group(2))])
-
+            for regexp in self.regexes:
+                match = regexp.match(line)
+                if match:
+                    result.append([float(match.group('time')), float(match.group('value'))])
+                    break # only match one regexp per line
         return result
 
+class PingRunner(RegexpRunner):
+    """Runner for ping/ping6 in timestamped (-D) mode."""
+
+    regexes = [re.compile(r'^\[(?P<time>[0-9]+\.[0-9]+)\].*time=(?P<value>[0-9]+(?:\.[0-9]+)?) ms$'),
+               re.compile(r'^\[(?P<time>[0-9]+\.[0-9]+)\].*:.*, (?P<value>[0-9]+(?:\.[0-9]+)?) ms \(.*\)$')]
+
+class HttpGetterRunner(RegexpRunner):
+
+    regexes = [re.compile(r'^\[(?P<time>[0-9]+\.[0-9]+)\].*in (?P<value>[0-9]+(?:\.[0-9]+)?) seconds.$')]
+
 class IperfCsvRunner(ProcessRunner):
     """Runner for iperf csv output (-y C), possibly with unix timestamp patch."""
 
@@ -271,10 +425,28 @@ class TcRunner(ProcessRunner):
                 sys.stderr.write("Warning: Missing value for %s" % key)
         return result
 
+class NullRunner(object):
+    def __init__(self, *args, **kwargs):
+        self.result = None
+        self.command = 'null'
+        self.returncode = 0
+        self.out = self.err = ''
+    # Emulate threading interface to fit into aggregator usage.
+    def start(self):
+        pass
+    def join(self):
+        pass
+    def isAlive(self):
+        return False
+    def kill(self):
+        pass
+
 class ComputingRunner(object):
     command = "Computed"
-    def __init__(self, name, apply_to=None, *args, **kwargs):
+    supported_meta = ['MEAN_VALUE']
+    def __init__(self, name, settings, apply_to=None, *args, **kwargs):
         self.name = name
+        self.settings = settings
         if apply_to is None:
             self.keys = []
         else:
@@ -309,6 +481,16 @@ class ComputingRunner(object):
             else:
                 new_res.append(self.compute(values))
 
+        meta = res.meta('SERIES_META')
+        meta[self.name] = {}
+        for mk in self.supported_meta:
+            vals = []
+            for k in keys:
+                if k in meta and mk in meta[k]:
+                    vals.append(meta[k][mk])
+            if vals:
+                meta[self.name][mk] = self.compute(vals)
+
         res.add_result(self.name, new_res)
         return res
 
@@ -340,3 +522,24 @@ class SumRunner(ComputingRunner):
     command = "Sum (computed)"
     def compute(self,values):
         return math.fsum(values)
+
+class DiffMinRunner(ComputingRunner):
+    command = "Diff from min (computed)"
+    def result(self, res):
+        if not self.keys:
+            return res
+
+        key = self.keys[0]
+
+        data = [i for i in res[key] if i is not None]
+        if not data:
+            res.add_result(self.name, [None]*len(res[key]))
+        else:
+            min_val = min(data)
+            res.add_result(self.name, [i-min_val if i is not None else None for i in res[key]])
+        return res
+
+class FairnessRunner(ComputingRunner):
+    command = "Fairness (computed)"
+    def compute(self,values):
+        return math.fsum(values)**2/(len(values)*math.fsum([x**2 for x in values]))
